"""Module which defines and handles Glia Cells and connectivity"""

import logging
from pathlib import Path

import libsonata
import numpy as np

from .cell_distributor import CellDistributor
from .connection import Connection
from .connection_manager import ConnectionManagerBase
from .core import (
    MPI,
    EngineBase,
    NeuronWrapper as Nd,
    mpi_no_errors,
)
from .core.configuration import ConfigurationError, GlobalConfig, LogLevel
from .io.sonata_config import ConnectionTypes
from .io.synapse_reader import SonataReader, SynapseParameters
from .metype import BaseCell
from .morphio_wrapper import MorphIOWrapper
from .utils.pyutils import append_recarray


class Astrocyte(BaseCell):
    __slots__ = ("_gluts", "glut_soma", "is_resized_sections_warning", "section_names")

    def __init__(self, gid, meinfos, circuit_conf):
        """Initialize an Astrocyte cell:

        - create the cell from Cell.hoc
        - add the morphology
        - resize sections if necessary
        - add cadifus, GlutReceive (they will be connected later)
        - record section_names for creating connections later
        """
        super().__init__()

        # Create the cell
        self._cellref = Nd.Cell(gid)  # cell instantiated with Cell.hoc
        # load and apply morphology
        morph = MorphIOWrapper(
            Path(circuit_conf.MorphologyPath)
            / f"{meinfos.morph_name}.{circuit_conf.MorphologyType}"
        )
        self._cellref.AddHocMorph(morph.morph_as_hoc())
        # Recalculate number of segments and sections
        self._cellref.geom_nseg_fixed()
        self._cellref.geom_nsec()

        logging.debug("Instantiating NGV cell gid=%d", gid)

        # assigned later
        self._gluts = {}
        self.is_resized_sections_warning = False
        for sec in self.all:
            self._init_basic_section(sec)

        # add GlutReceiveSoma (only for metabolism)
        soma = self._cellref.soma[0]
        self.glut_soma = Nd.GlutReceiveSoma(soma(0.5), sec=soma)

        self.gid = gid

    def _init_basic_section(self, sec):
        """Initialize a basic NEURON section with standard mechanisms.

        This function ensures the section has a single compartment (`nseg = 1`),
        resizing it if necessary. After, it inserts the 'cadifus' mechanism used
        for calcium diffusion. It returns whether the section had more
        than one segment before resizing.

        Parameters:
            sec (neuron.h.Section): The section to initialize.
        """
        # resize if necessary
        if sec.nseg > 1:
            self.is_resized_sections_warning = True
            sec.nseg = 1
        # add cadifus mechanism for calcium diffusion
        sec.insert("cadifus")

    def _init_endfoot_section(
        self,
        sec,
        parent_id: int,
        length: float,
        diameter: float,
        R0pas: float,  # noqa: N803
    ) -> bool:
        """Initialize an endfoot NEURON section with custom geometry and mechanisms.

        Parameters:
            sec (neuron.h.Section): The endfoot section to initialize.
            parent_id (int): Index to identify the parent section.
            length (float): Length of the endfoot section.
            diameter (float): Diameter of the endfoot section.
            R0pas (float): Passive resistance parameter for the vascouplingB mechanism.
        """
        self._init_basic_section(sec)
        sec.L = length
        sec.diam = diameter
        sec.insert("vascouplingB")
        sec(0.5).vascouplingB.R0pas = R0pas
        # connect to parent sec
        parent_sec = self.get_sec(parent_id + 1)
        sec.connect(parent_sec)

    def get_glut(self, section_id):
        """Return cached GlutReceive object for a section, creating it if needed."""
        if section_id in self._gluts:
            return self._gluts[section_id]
        sec = self.get_sec(section_id)
        glut = Nd.GlutReceive(sec(0.5), sec=sec)
        sec(0.5).cadifus._ref_glu2 = glut._ref_glut
        self._gluts[section_id] = glut
        return glut

    @property
    def gid(self) -> int:
        """Get the gid as an integer."""
        return int(self._cellref.gid)

    @gid.setter
    def gid(self, val: int):
        """Set the gid value."""
        self._cellref.gid = val

    @property
    def all(self):
        """Return the main SectionList (`_cellref.all`).

        Note:
            This list does **not** include all sections. Specifically,
            endfeet sections are excluded and can be accessed via the `endfeet` attribute.
        """
        return self._cellref.all

    @property
    def endfeet(self):
        """Returns _cellref.endfeet (SectionList)."""
        if hasattr(self._cellref, "endfeet") and self._cellref.endfeet is not None:
            return self._cellref.endfeet
        return Nd.SectionList()

    def add_endfeet(self, parent_ids, lengths, diameters, R0passes):  # noqa: N803
        assert len(parent_ids) == len(lengths) == len(diameters) == len(R0passes)
        self._cellref.execute_commands(
            [
                f"create endfoot[{len(parent_ids)}]",
                "endfeet = new SectionList()",
                'forsec "endfoot" endfeet.append',
            ]
        )
        for sec, parent_id, length, diameter, R0pas in zip(
            self.endfeet, parent_ids, lengths, diameters, R0passes
        ):
            self._init_endfoot_section(sec, parent_id, length, diameter, R0pas)

    @property
    def glut_list(self) -> list:
        # necessary for legacy compatibility with metabolism
        return [*self._gluts.values(), self.glut_soma]

    def connect2target(self, target_pp=None):
        return Nd.NetCon(self._cellref.soma[0](1)._ref_v, target_pp, sec=self._cellref.soma[0])

    @staticmethod
    def getThreshold():
        return 0.114648


class AstrocyteManager(CellDistributor):
    """Manages Astrocyte cells, extending CellDistributor

    Behaves like CellDistributor but uses the Astrocyte cell type.
    The difference lies only in the Cell Type.
    """

    CellType = Astrocyte
    _sonata_with_extra_attrs = False

    def _emit_resized_section_warnings(self):
        """Collect and emit warnings for cells that had sections resized."""
        gids = [cell.gid for cell in self._gid2cell.values() if cell.is_resized_sections_warning]
        # Gather all warning maps to rank 0
        gids = MPI.py_gather(gids, 0)

        if MPI.rank == 0:
            # flatten the list of gids
            gids = [gid for sublist in gids for gid in sublist]
            gids = sorted(gids)
            logging.warning(
                "Cells %s emitted warning: '%s'",
                sorted(gids),
                "Multi-compartment sections are not allowed at the moment."
                "Resizing to 1 compartment.",
            )

    @mpi_no_errors
    def _instantiate_cells(self, cell_type=None, **_opts):
        super()._instantiate_cells(cell_type=cell_type, **_opts)
        self._emit_resized_section_warnings()

    @mpi_no_errors
    def _instantiate_cells_dry(self, cell_type, skip_metypes, **_opts):
        super()._instantiate_cells_dry(cell_type=cell_type, skip_metypes=skip_metypes, **_opts)
        self._emit_resized_section_warnings()


class NeuroGliaConnParameters(SynapseParameters):
    """Neuron-to-glia connection parameters.

    This class overrides the `_fields` attribute from `SynapseParameters` to define
    parameters specific to neuro-glial interactions.

    The `_optional` and `_reserved` dictionaries are inherited unchanged from the base class.

    Note:
        - Only `_fields` is overridden.
        - All methods and behavior are reused from the base class.
    """

    _fields = {
        "tgid": np.int64,
        "synapse_id": np.int64,
        "astrocyte_section_id": np.int64,
        "astrocyte_segment_id": np.int64,
        "astrocyte_segment_offset": np.float64,
    }


class NeuroGlialSynapseReader(SonataReader):
    LOOKUP_BY_TARGET_IDS = False
    Parameters = NeuroGliaConnParameters
    custom_parameters = set()


class NeuroGlialConnection(Connection):
    neurons_not_found = set()
    neurons_attached = set()
    netcon_delay = 0.05
    syn_gid_offset = 1_000_000  # Below 1M is reserved for cell ids

    def add_synapse(self, syn_tpoints, params_obj, syn_id=None):
        # Only store params. Glia have mechanisms pre-created
        self._synapse_params = append_recarray(self._synapse_params, params_obj)

    def finalize(self, astrocyte, base_seed, *, base_connections=None, **kw):
        """Bind each glia connection to synapses in connections target cells via
        the assigned unique gid.
        """
        self._netcons = []
        pc = Nd.pc

        if GlobalConfig.debug_conn:
            if GlobalConfig.debug_conn == [self.tgid]:
                logging.debug("Finalizing conn %s. N params: %d", self, len(self._synapse_params))
            elif GlobalConfig.debug_conn == [self.sgid, self.tgid]:
                logging.debug("Finalizing conn %s. Params:\n%s", self, self._synapse_params)

        for syn_params in self._synapse_params:
            syn_gid = self.syn_gid_offset + syn_params.synapse_id

            # netcon to GlutReceive
            glut_idx = int(syn_params.astrocyte_section_id)
            glut_obj = astrocyte.get_glut(glut_idx)
            netcon = pc.gid_connect(syn_gid, glut_obj)
            netcon.delay = self.netcon_delay
            self._netcons.append(netcon)

            # netcon to GlutReceiveSoma for metabolism
            logging.debug("[NGV] Conn %s linking synapse id %d to Astrocyte", self, syn_gid)
            netcon = pc.gid_connect(syn_gid, astrocyte.glut_soma)
            netcon.delay = self.netcon_delay
            self._netcons.append(netcon)

        return len(self._synapse_params)


class NeuroGliaConnManager(ConnectionManagerBase):
    """A Connection Manager for Neuro-Glia connections

    NOTE: We assume the only kind of connections for Glia are Neuron-Glia
    If one day Astrocytes have connections among themselves a sub ConnectionManager
    must be used
    """

    ustate_netcon_threshold = 0.0
    ustate_netcon_delay = 0.0
    ustate_netcon_weight = 1.1

    CONNECTIONS_TYPE = ConnectionTypes.NeuroGlial
    conn_factory = NeuroGlialConnection
    SynapseReader = NeuroGlialSynapseReader

    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        kw.pop("load_offsets")
        super().__init__(circuit_conf, target_manager, cell_manager, src_cell_manager, **kw)

    @staticmethod
    def _add_synapses(cur_conn, syns_params, _syn_type_restrict=None, _base_id=0):
        for syn_params in syns_params:
            cur_conn.add_synapse(None, syn_params)

    def finalize(self, base_seed=0, *_):
        """Instantiate connections to the simulator.

        This is a two-step process:
        First we create netcons to listen events on target synapses.Ustate,
        and assign them a virtual gid.
        Second, as part of NeuroGlialConnection.finalize(), we attach netcons to
        the target glia cell, listening for "signals" from the virtual gids.
        """
        logging.info("Creating virtual cells on target Neurons for coupling to GLIA...")
        base_manager = next(self._src_cell_manager.connection_managers.values())

        total_created = self._create_synapse_ustate_endpoints(base_manager)

        logging.info("(RANK 0) Created %d Virtual GIDs for synapses.", total_created)

        super().finalize(
            base_seed,
            base_connections=None,
            conn_type="NeuronGlia connections",
        )

        if NeuroGlialConnection.neurons_not_found:
            logging.warning(
                "Missing cells to couple Glia to: %d", len(NeuroGlialConnection.neurons_not_found)
            )

    @staticmethod
    def _create_synapse_ustate_endpoints(base_manager):
        """Creating an endpoint netcon to listen for events in synapse.Ustate
        Netcon ids are directly the synapse id (hence we are limited in number space)

        Note: we assume that the source synapse has a Ustate variable
        """
        pc = Nd.pc
        syn_gid_base = NeuroGlialConnection.syn_gid_offset
        total_created = 0

        for conn in base_manager.all_connections():
            syn_objs = conn.synapses
            tgid_syn_offset = syn_gid_base + conn.synapses_offset
            logging.debug("Tgid: %d, Base syn offset: %d", conn.tgid, tgid_syn_offset)

            for param_i, sec in conn.sections_with_synapses:
                if conn.synapse_params[param_i].synType < 100:  # Skip Inhibitory
                    continue
                synapse_gid = tgid_syn_offset + param_i
                pc.set_gid2node(synapse_gid, MPI.rank)
                netcon = Nd.NetCon(
                    syn_objs[param_i]._ref_Ustate,
                    None,
                    NeuroGliaConnManager.ustate_netcon_threshold,
                    NeuroGliaConnManager.ustate_netcon_delay,
                    NeuroGliaConnManager.ustate_netcon_weight,
                    sec=sec,
                )
                # set the v-gid (that is actually an synapse id) to the netcon. Useful for
                # reporting and debugging
                pc.cell(synapse_gid, netcon)
                if GlobalConfig.verbosity >= LogLevel.DEBUG:
                    netcon.record(
                        lambda tgid=conn.tgid, synapse_gid=synapse_gid: print(  # noqa: T201
                            f"[gid={tgid}] Ustate netcon event. Spiking via v-gid={synapse_gid}"
                        )
                    )

                conn._netcons.append(netcon)
                total_created += 1

        return total_created


class GlioVascularManager(ConnectionManagerBase):
    CONNECTIONS_TYPE = ConnectionTypes.GlioVascular
    InnerConnectivityCls = None  # No synapses

    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        if cell_manager.circuit_target is None:
            raise ConfigurationError("Circuit target is required for GlioVascular projections")
        if "Path" not in circuit_conf:
            raise ConfigurationError("Missing GlioVascular Sonata file via 'Path' configuration")
        if "VasculaturePath" not in circuit_conf:
            raise ConfigurationError(
                "Missing Vasculature Sonata file via 'VasculaturePath' configuration"
            )

        super().__init__(circuit_conf, target_manager, cell_manager, src_cell_manager, **kw)
        self._astro_ids = self._cell_manager.local_nodes.gids(raw_gids=True)
        self._gid_offset = self._cell_manager.local_nodes.offset

    def open_edge_location(self, sonata_source, circuit_conf, **__):
        logging.info("GlioVascular sonata file %s", sonata_source)
        # sonata files can have multiple populations. In building we only use one
        # per file, hence this two lines below to access the first and only pop in
        # the file
        edge_file, *pop = sonata_source.split(":")
        storage = libsonata.EdgeStorage(edge_file)
        pop_name = pop[0] if pop else next(iter(storage.population_names))
        self._gliovascular = storage.open_population(pop_name)

        storage = libsonata.NodeStorage(circuit_conf["VasculaturePath"])
        pop_name = next(iter(storage.population_names))
        self._vasculature = storage.open_population(pop_name)

    def create_connections(self, *_, **__):
        # it also creates endfeet
        logging.info("Creating GlioVascular virtual connections")
        # Retrieve endfeet selections for GLIA gids on the current processor

        for astro_id in self._astro_ids:
            self._connect_endfeet(astro_id)

    def _connect_endfeet(self, astro_id):
        endfeet = self._gliovascular.afferent_edges(astro_id - 1)  # 0-based for libsonata API
        if endfeet.flat_size > 0:
            # Get endfeet input

            parent_section_ids = self._gliovascular.get_attribute("astrocyte_section_id", endfeet)
            lengths = self._gliovascular.get_attribute("endfoot_compartment_length", endfeet)
            diameters = self._gliovascular.get_attribute("endfoot_compartment_diameter", endfeet)

            # Retrieve instantiated astrocyte
            astrocyte = self._cell_manager.gid2cell[astro_id + self._gid_offset]

            # Retrieve R0pas
            vasc_node_ids = libsonata.Selection(self._gliovascular.source_nodes(endfeet))
            d_vessel_starts = self._vasculature.get_attribute("start_diameter", vasc_node_ids)
            d_vessel_ends = self._vasculature.get_attribute("end_diameter", vasc_node_ids)
            R0passes = (d_vessel_starts + d_vessel_ends) / 4

            astrocyte.add_endfeet(parent_section_ids, lengths, diameters, R0passes)

    def finalize(self, *_, **__):
        pass  # No synpases/netcons


class NGVEngine(EngineBase):
    CellManagerCls = AstrocyteManager
    ConnectionTypes = {
        ConnectionTypes.NeuroGlial: NeuroGliaConnManager,
        ConnectionTypes.GlioVascular: GlioVascularManager,
    }



"""Module to load configuration from a libsonata config"""

import json
import logging
import os.path
from enum import Enum

import libsonata


class ConnectionTypes(str, Enum):
    Synaptic = "Synaptic"
    GapJunction = "GapJunction"
    NeuroModulation = "NeuroModulation"
    NeuroGlial = "NeuroGlial"
    GlioVascular = "GlioVascular"


class SonataConfig:
    __slots__ = (
        "_circuit_conf",  # libsonata.CircuitConfig
        "_circuits",
        "_sim_conf",  # libsonata.SimulationConfig
        # Currently, the `inputs` of a simulation_config is a json object,
        # which is unordered; however, the order of stimuli matter, so try and
        # recover the order defined in the json file: this assumes that `json.load`
        # keeps it; *which is not guaranteed* see the discussion:
        # https://github.com/openbraininstitute/neurodamus/issues/217#issuecomment-2827930163
        # the SONATA specification should be updated to be a list
        "_stable_inputs_order",
    )

    def __init__(self, config_path):
        self._sim_conf = libsonata.SimulationConfig.from_file(config_path)

        with open(config_path, encoding="utf-8") as fd:
            if inputs := json.load(fd).get("inputs", None):
                self._stable_inputs_order = tuple(inputs.keys())
            else:
                self._stable_inputs_order = ()

        self._circuit_conf = libsonata.CircuitConfig.from_file(self._sim_conf.network)
        self._circuits = self._extract_circuits_info()

    @property
    def beta_features(self):
        return self._sim_conf.beta_features

    @property
    def parsedRun(self):
        item_translation = {
            # Mandatory
            "tstop": "Duration",
            "dt": "Dt",
            "random_seed": "BaseSeed",
            # Optional
            "tstart": "Start",
            "spike_threshold": "SpikeThreshold",
            "integration_method": "SecondOrder",
            "electrodes_file": "LFPWeightsPath",
        }
        parsed_run = self._translate_dict(item_translation, self._sim_conf.run)

        self._adapt_libsonata_fields(parsed_run)

        # "OutputRoot" and "SpikesFile" will be read from self._sim_conf.output
        # once libsonata resolves the manifest info
        parsed_run["OutputRoot"] = self._sim_conf.output.output_dir
        parsed_run["config_node_sets_file"] = self._circuit_conf.node_sets_path
        parsed_run["TargetFile"] = self._circuit_conf.node_sets_path
        parsed_run["SpikesFile"] = self._sim_conf.output.spikes_file
        parsed_run["SpikesSortOrder"] = self._sim_conf.output.spikes_sort_order.name
        parsed_run["Simulator"] = self._sim_conf.target_simulator.name
        parsed_run["TargetFile"] = self._sim_conf.node_sets_file
        parsed_run["CircuitTarget"] = self._sim_conf.node_set
        parsed_run["Celsius"] = self._sim_conf.conditions.celsius
        parsed_run["V_Init"] = self._sim_conf.conditions.v_init
        parsed_run["ExtracellularCalcium"] = self._sim_conf.conditions.extracellular_calcium
        parsed_run["SpikeLocation"] = self._sim_conf.conditions.spike_location.name
        parsed_run["compartment_sets_file"] = self._sim_conf.compartment_sets_file
        return parsed_run

    @property
    def Conditions(self):
        item_translation = {"randomize_gaba_rise_time": "randomize_Gaba_risetime"}
        conditions = {}
        blacklist = (
            "Celsius",
            "VInit",
            "ExtracellularCalcium",
            "ListModificationNames",
            "SpikeLocation",
        )
        for key, value in self._translate_dict(item_translation, self._sim_conf.conditions).items():
            if key in blacklist:
                continue
            if key == "Mechanisms":
                for suffix, dict_var in value.items():
                    for name, val in dict_var.items():
                        conditions[name + "_" + suffix] = val
            else:
                conditions[key] = value

        conditions["randomize_Gaba_risetime"] = str(conditions["randomize_Gaba_risetime"])

        return {"Conditions": conditions}

    def _extract_circuits_info(self) -> dict:  # noqa: C901
        """Extract the circuits information from confile file with libsonata.CircuitConfig parser,
        return a dictionary of circuit info as:
        {
            pop_name: { "CellLibraryFile": ...,
                        "CircuitTarget": ...,
                        "MorphologyPath": ...,
                        "MorphologyType": ...,
                        "METypePath": ...,
                        "Engine": ...,
                        "nrnPath": ...,
                        "PopulationType": ...
                }
        }
        It will be used to build the internal circuit structure CircuitConfig in configuration.py
        """
        node_info_to_circuit = {"nodes_file": "CellLibraryFile", "type": "PopulationType"}

        simulation_nodeset_name = self._sim_conf.node_set or ""
        if not simulation_nodeset_name:
            logging.warning("Simulating all populations from all node files...")

        network = json.loads(self._circuit_conf.expanded_json)["networks"]

        def make_circuit(nodes_file, node_pop_name, population_info):
            if not os.path.isabs(nodes_file):
                nodes_file = os.path.join(os.path.dirname(self._sim_conf.network), nodes_file)
            circuit_config = dict(
                CellLibraryFile=nodes_file,
                # Use the extended ":" syntax to filter the nodeset by the related population
                CircuitTarget=node_pop_name + ":" + simulation_nodeset_name,
                **{
                    node_info_to_circuit.get(key, key): value
                    for key, value in population_info.items()
                },
            )
            node_prop = self._circuit_conf.node_population_properties(node_pop_name)
            circuit_config["MorphologyPath"] = node_prop.morphologies_dir
            circuit_config["MorphologyType"] = "h5" if node_prop.type == "astrocyte" else "swc"
            circuit_config["METypePath"] = node_prop.biophysical_neuron_models_dir
            if node_prop.alternate_morphology_formats:
                if "neurolucida-asc" in node_prop.alternate_morphology_formats:
                    circuit_config["MorphologyPath"] = node_prop.alternate_morphology_formats[
                        "neurolucida-asc"
                    ]
                    circuit_config["MorphologyType"] = "asc"
                elif "h5v1" in node_prop.alternate_morphology_formats:
                    circuit_config["MorphologyPath"] = node_prop.alternate_morphology_formats[
                        "h5v1"
                    ]
                    circuit_config["MorphologyType"] = "h5"
            circuit_config["Engine"] = "NGV" if node_prop.type == "astrocyte" else "METype"

            # Find inner connectivity
            # NOTE: Inner connectivity is a special kind of projection, and represents the circuit
            # default set of connections. Even though nowadays we can potentially consider
            # all connectivity as projections, under certain circuitry, like NGV, order matters and
            # therefore we keep inner connectivity to ensure they are created in the same order,
            # respecting engine precedence
            # For edges to be considered inner connectivity they must be named "default"
            for edge_config in network.get("edges") or []:
                if "nrnPath" in circuit_config:
                    break  # Already found

                for edge_pop_name in edge_config["populations"]:
                    edge_storage = self._circuit_conf.edge_population(edge_pop_name)
                    edge_type = self._circuit_conf.edge_population_properties(edge_pop_name).type
                    inner_pop_name = f"{node_pop_name}__{node_pop_name}__chemical"
                    if edge_pop_name == inner_pop_name or (
                        edge_storage.source == edge_storage.target == node_pop_name
                        and edge_type == "chemical"
                        and edge_pop_name == "default"
                    ):
                        edges_file = edge_config["edges_file"]
                        if not os.path.isabs(edges_file):
                            edges_file = os.path.join(
                                os.path.dirname(self._sim_conf.network), edges_file
                            )
                        edge_pop_path = edges_file + ":" + edge_pop_name
                        circuit_config["nrnPath"] = edge_pop_path
                        break

            circuit_config.setdefault("nrnPath", False)
            logging.debug("Circuit config for node pop '%s': %s", node_pop_name, circuit_config)
            return circuit_config

        return {
            pop_name: make_circuit(node_file_info["nodes_file"], pop_name, pop_info)
            for node_file_info in network["nodes"]
            for pop_name, pop_info in node_file_info["populations"].items()
            if pop_info.get("type") != "vasculature"
        }

    @property
    def Circuit(self):
        return self._circuits

    @property
    def parsedProjections(self):
        projection_type_convert = {
            "chemical": ConnectionTypes.Synaptic,
            "electrical": ConnectionTypes.GapJunction,
            "synapse_astrocyte": ConnectionTypes.NeuroGlial,
            "endfoot": ConnectionTypes.GlioVascular,
            "neuromodulatory": ConnectionTypes.NeuroModulation,
        }
        internal_edge_pops = {c_conf["nrnPath"] for c_conf in self._circuits.values()}
        projections = {}

        networks = json.loads(self._circuit_conf.expanded_json)["networks"]
        for edge_config in networks.get("edges") or []:
            for edge_pop_name, edge_pop_config in edge_config["populations"].items():
                edge_pop = self._circuit_conf.edge_population(edge_pop_name)
                pop_type = edge_pop_config.get("type", "chemical")

                if pop_type not in projection_type_convert:
                    logging.warning("Unhandled synapse type: %s", pop_type)
                    continue

                edges_file = edge_config["edges_file"]
                if not os.path.isabs(edges_file):
                    edges_file = os.path.join(os.path.dirname(self._sim_conf.network), edges_file)

                # skip inner connectivity populations
                edge_pop_path = edges_file + ":" + edge_pop_name
                if edge_pop_path in internal_edge_pops:
                    continue

                projection = {
                    "Path": edge_pop_path,
                    "Source": edge_pop.source + ":",
                    "Destination": edge_pop.target + ":",
                    "Type": projection_type_convert.get(pop_type),
                }
                # Reverse projection direction for Astrocyte projection: from neurons to astrocytes
                if projection["Type"] == ConnectionTypes.NeuroGlial:
                    projection["Source"], projection["Destination"] = (
                        projection["Destination"],
                        projection["Source"],
                    )
                elif projection["Type"] == ConnectionTypes.GlioVascular:
                    vasculature_popnames = [
                        name
                        for node_info in networks["nodes"]
                        for name, pop_info in node_info["populations"].items()
                        if pop_info.get("type") == "vasculature"
                    ]
                    if vasculature_popnames:
                        assert len(vasculature_popnames) == 1
                        projection["VasculaturePath"] = (
                            self._circuit_conf.node_population_properties(
                                vasculature_popnames[0]
                            ).elements_path
                        )

                proj_name = f"{edge_pop_name}__{edge_pop.source}-{edge_pop.target}"
                projections[proj_name] = projection

        return projections

    @property
    def parsedConnects(self):
        item_translation = {
            "target": "Destination",
            "modoverride": "ModOverride",
            "synapse_delay_override": "SynDelayOverride",
            "neuromodulation_dtc": "NeuromodDtc",
            "neuromodulation_strength": "NeuromodStrength",
        }
        connects = {
            libsonata_conn.name: self._translate_dict(item_translation, libsonata_conn)
            for libsonata_conn in self._sim_conf.connection_overrides()
        }
        return connects

    @property
    def parsedStimuli(self) -> list:
        """Read the inputs information parsed by libsonata,
        and convert them to the internal parameters used by StimulusManager
        """
        item_translation = {
            "module": "Pattern",
            "input_type": "Mode",
            "random_seed": "Seed",
            "series_resistance": "RS",
            "node_set": "Target",
            "sd_percent": "SDPercent",
            "relative_skew": "RelativeSkew",
        }
        input_type_translation = {
            "spikes": "Current",
            "current_clamp": "Current",
            "voltage_clamp": "Voltage",
            "extracellular_stimulation": "Extracellular",
            "conductance": "Conductance",
        }
        module_translation = {"seclamp": "SEClamp", "subthreshold": "SubThreshold"}

        stimuli = []
        # TODO: loop over self._sim_conf.inputs() list after updating SONATA SPEC and libsonata API,
        # The order of stimulus injection could lead to minor difference on the results
        # so need to preserve it as in the config file
        for name in self._stable_inputs_order:
            stimulus = self._translate_dict(item_translation, self._sim_conf.input(name))
            self._adapt_libsonata_fields(stimulus)
            stimulus["Pattern"] = module_translation.get(
                stimulus["Pattern"], snake_to_camel(stimulus["Pattern"])
            )
            stimulus["Mode"] = input_type_translation.get(stimulus["Mode"], stimulus["Mode"])
            stimulus["Name"] = name
            stimuli.append(stimulus)

        return stimuli

    @property
    def parsedReports(self):
        item_translation = {
            "type": "Type",
            "cells": "Target",
            "sections": "Sections",
            "scaling": "Scaling",
            "compartments": "Compartments",
            "variable_name": "ReportOn",
            "unit": "Unit",
            "dt": "Dt",
            "start_time": "StartTime",
            "end_time": "EndTime",
            "file_name": "FileName",
            "enabled": "Enabled",
        }
        reports = {}
        for name in self._sim_conf.list_report_names:
            rep = self._translate_dict(item_translation, self._sim_conf.report(name))
            # Adapt enums and variable names read from libsonata
            self._adapt_libsonata_fields(rep)
            # Format is SONATA with sonata_config
            rep["Format"] = "SONATA"
            reports[name] = rep
            rep["Scaling"] = snake_to_camel(rep["Scaling"])

        return reports

    @property
    def parsedModifications(self):
        item_translation = {"node_set": "Target"}
        result = {}
        for modification in self._sim_conf.conditions.modifications():
            setting = self._translate_dict(item_translation, modification)
            self._adapt_libsonata_fields(setting)
            result[modification.name] = setting

        return result

    @staticmethod
    def _adapt_libsonata_fields(rep):
        for key in rep:
            # Convert enums to its string representation
            if key in {
                "Type",
                "Sections",
                "Scaling",
                "Compartments",
                "Mode",
                "Pattern",
                "SpikeLocation",
            } and not isinstance(rep[key], str):
                rep[key] = rep[key].name
            # Convert comma separated variable names to space separated
            if key == "ReportOn":
                rep[key] = rep[key].replace(",", " ")
            # Get the int value of the enum
            elif key == "SecondOrder":
                rep[key] = int(rep[key])

    @staticmethod
    def _translate_dict(item_translation, libsonata_obj) -> dict:
        """Translate SONATA/libsonata key names (snake_case) to Neurodamus internal paramters"""
        attrs = [
            x
            for x in dir(libsonata_obj)
            if not x.startswith("__") and not callable(getattr(libsonata_obj, x))
        ]
        result = {}
        for att in attrs:
            key = item_translation.get(att, snake_to_camel(att))
            parsed_value = getattr(libsonata_obj, att)
            if parsed_value is not None:
                result[key] = parsed_value
        return result


def snake_to_camel(word):
    return "".join(x.capitalize() or "_" for x in word.split("_"))



"""Loggeers init & Formatters"""

import logging as _logging
import os
import sys

from .pyutils import ConsoleColors

STAGE_LOGLEVEL = 25
VERBOSE_LOGLEVEL = 15
ALWAYS_LEVEL = 40

# New log levels, for stage (minimal output, verbose)
_logging.addLevelName(STAGE_LOGLEVEL, "STEP")
_logging.STAGE = STAGE_LOGLEVEL
_logging.addLevelName(VERBOSE_LOGLEVEL, "VERB")
_logging.VERBOSE = VERBOSE_LOGLEVEL


def log_stage(msg, *args):
    """Shortcut to log a messge with the STAGE level"""
    _logging.log(STAGE_LOGLEVEL, msg, *args)


def log_verbose(msg, *args):
    """Shortcut to log a messge with the VERBOSE level"""
    _logging.log(VERBOSE_LOGLEVEL, msg, *args)


def log_all(level, msg, *args):
    """Like logging.log, but always displays. Level is used for style only"""
    _logging.log(ALWAYS_LEVEL, msg, *args, extra={"ulevel": level})


class _LevelColorFormatter(_logging.Formatter):
    COLORS = {
        _logging.CRITICAL: ConsoleColors.RED,
        _logging.ERROR: ConsoleColors.RED,
        _logging.WARNING: ConsoleColors.YELLOW,
        STAGE_LOGLEVEL: ConsoleColors.DEFAULT + ConsoleColors.BOLD,
        _logging.INFO: ConsoleColors.BLUE,
        VERBOSE_LOGLEVEL: ConsoleColors.BLUE,
        _logging.DEBUG: ConsoleColors.DEFAULT + ConsoleColors.DIM,
    }

    _logfmt = "[%(levelname)s] %(message)s"
    _datefmt = "%b.%d %H:%M:%S"
    _level_tabs = {VERBOSE_LOGLEVEL: " -> ", _logging.DEBUG: " + "}

    def __init__(self, with_time=True, rank=None, use_color=True, **kw):
        super().__init__(self._logfmt, self._datefmt, **kw)
        self._rank = rank
        self._with_time = with_time
        self._use_color = use_color

    def format(self, record):
        if hasattr(record, "ulevel"):
            record.levelno = record.ulevel
            record.levelname = _logging.getLevelName(record.levelno)
        style = self.COLORS.get(record.levelno)
        if style is not None:
            record.levelname = self._format_level(record, style)
            record.msg = self._format_msg(record, style)
        return super().format(record)

    def _format_level(self, record, style):
        if not self._use_color:
            return record.levelname
        return ConsoleColors.format_text(record.levelname, style)

    def _format_msg(self, record, style):
        msg = ""
        if self._with_time:
            msg += f"({self.formatTime(record, self._datefmt)}) " + msg
        # Show rank only for ERRORs
        if self._rank is not None and record.levelno >= _logging.ERROR:
            msg += f"(rank {self._rank:d}) "

        levelno = record.levelno
        msg += self._level_tabs.get(levelno, "") + record.msg

        if not self._use_color:
            return msg
        return (
            ConsoleColors.format_text(msg, style)
            if levelno >= _logging.WARNING or levelno == VERBOSE_LOGLEVEL
            else ConsoleColors.format_text(msg, ConsoleColors.DEFAULT, style)
        )


def setup_logging(loglevel, logfile=None, rank=None, use_color=True):
    """Setup neurodamus logging.
    Features tabs and colors output to stdout and pydamus.log

    Args:
      loglevel (int): minimum loglevel for emitting messages
      logfile: The destination for log messages besides stdout
      rank: A tag so warnings/errors are correctly identified in case of MPI
    """
    if getattr(setup_logging, "logging_initted", False):
        return
    assert isinstance(loglevel, int)
    loglevel = min(loglevel, 3)

    verbosity_levels = [
        _logging.WARNING,  # pos 0: Minimum possible logging level
        _logging.INFO,
        _logging.VERBOSE,
        _logging.DEBUG,
    ]

    # stdout
    hdlr = _logging.StreamHandler(sys.stdout)
    if use_color:
        if os.environ.get("ENVIRONMENT") == "BATCH":
            use_color = False
        else:
            try:
                sys.stdout.tell()  # works only if it's file
                use_color = False
            except OSError:
                pass
    hdlr.setFormatter(_LevelColorFormatter(with_time=False, rank=rank, use_color=use_color))
    if rank == 0:
        _logging.root.setLevel(verbosity_levels[loglevel])
    else:
        _logging.root.setLevel(_logging.ERROR)

    del _logging.root.handlers[:]
    _logging.root.addHandler(hdlr)

    if logfile:
        fileh = _logging.FileHandler(logfile, encoding="utf-8")
        fileh.setFormatter(_LevelColorFormatter(with_time=True, rank=rank, use_color=False))
        _logging.root.addHandler(fileh)

    setup_logging.logging_initted = True



# @file gap_junction_user_corrections.py
# @brief Script for loading user corrections on gap junction connectivity
# @author Oren Amsalem, Weina Ji
# @date 2024-09-09


import logging

from .core import MPI, NeuronWrapper as Nd
from .core.configuration import ConfigurationError, SimConfig

non_stochastic_mechs = [
    "NaTs2_t",
    "SKv3_1",
    "Nap_Et2",
    "Ih",
    "Im",
    "KdShu2007",
    "K_Pst",
    "K_Tst",
    "Ca",
    "SK_E2",
    "Ca_LVAst",
    "CaDynamics_E2",
    "NaTa_t",
    "CaDynamics_DC0",
    "Ca_HVA2",
    "NaTg",
    "TC_cad",
    "TC_ih_Bud97",
    "TC_Nap_Et2",
    "TC_iA",
    "TC_iL",
    "TC_HH",
    "TC_iT_Des98",
    "kdrb",
    "na3",
    "kap",
    "hd",
    "can",
    "cal",
    "cat",
    "cagk",
    "kca",
    "cacum",
    "kdb",
    "kmb",
    "kad",
    "nax",
    "cacumb",
]

stochastic_mechs = ["StochKv", "StochKv2", "StochKv3"]


def load_user_modifications(gj_manager):  # noqa: C901
    """Apply user modifications on gap junction connections, designed by @Oren Amsalem
    The modification parameters should be in the "beta_features" section of
    the simulation config file.
    """
    node_manager = gj_manager.cell_manager
    settings = SimConfig.beta_features
    gjc = settings.get("gjc")

    # deterministic_StochKv
    if settings.get("deterministic_stoch"):
        logging.info("Set deterministic = 1 for StochKv")
        _deterministic_stoch(node_manager)

    # update gap conductance
    if settings.get("procedure_type") in {"validation_sim", "find_holding_current"}:
        process_gap_conns = _update_conductance(gjc, gj_manager)
        all_ranks_total = int(MPI.py_sum(process_gap_conns, 0))
        logging.info("Set GJc = %s for %s gap synapses", gjc, all_ranks_total)

    # remove active channels
    remove_channels = settings.get("remove_channels")
    if remove_channels:
        if remove_channels == "all":
            rm_mechanisms = non_stochastic_mechs + stochastic_mechs
        elif remove_channels == "only_stoch":
            rm_mechanisms = stochastic_mechs
        elif remove_channels == "only_non_stoch":
            rm_mechanisms = non_stochastic_mechs
        else:
            logging.warning("Unknown GJ remove_channels setting: %s", remove_channels)
            rm_mechanisms = []
        if rm_mechanisms:
            logging.info("Remove channels type = %s", remove_channels)
            _perform_remove_channels(node_manager, rm_mechanisms)

    # load g_pas
    if filename := settings.get("load_g_pas_file"):
        processed_cells = _update_gpas(
            node_manager, filename, gjc, settings.get("correction_iteration_load", -1)
        )
        all_ranks_total = int(MPI.py_sum(processed_cells, 0))
        logging.info(
            "Update g_pas to fit %s - file %s for %s cells", gjc, filename, all_ranks_total
        )

    # load current clamps
    holding_ic_per_gid = {}
    if filename := settings.get("manual_MEComboInfo_file"):
        # Oren's note: If I manually injecting different holding current for each cell,
        # I will inject the current - the holding the emMEComboInfoFile
        if settings.get("procedure_type") == "find_holding_current":
            raise ConfigurationError("find_holding_current should not read manual_MEComboInfo_file")
        holding_ic_per_gid = _load_holding_ic(node_manager, filename, gjc=gjc)
        all_ranks_total = int(MPI.py_sum(len(holding_ic_per_gid), 0))
        logging.info(
            "Load holding_ic from manual_MEComboInfoFile %s for %s cells", filename, all_ranks_total
        )

    seclamp_per_gid = {}
    if settings.get("procedure_type") == "find_holding_current":
        seclamp_per_gid = _find_holding_current(node_manager, settings.get("vc_amp"))
        all_ranks_total = int(MPI.py_sum(len(seclamp_per_gid), 0))
        logging.info(
            "Inject holding voltages from file %s for %s cells",
            settings["vc_amp"],
            all_ranks_total,
        )

    return holding_ic_per_gid, seclamp_per_gid


def _update_conductance(gjc, gj_manager):
    """Update gap junction connection conductance"""
    n_conn = 0
    for conn in gj_manager.all_connections():
        conn.update_conductance(gjc)
        n_conn += 1
    return n_conn


def _deterministic_stoch(node_manager):
    """Enable deterministic_StochKV in cell"""
    for cell in node_manager.cells:
        for sec in cell._cellref.all:
            if "StochKv3" in dir(sec(0.5)):
                sec.deterministic_StochKv3 = 1
            if "StochKv2" in dir(sec(0.5)):
                sec.deterministic_StochKv2 = 1
            if "StochKv1" in dir(sec(0.5)):
                sec.deterministic_StochKv1 = 1


def _perform_remove_channels(node_manager, mechanisms: list):
    """Remove certain mechanisms from the cell"""
    for cell in node_manager.cells:
        for sec in cell._cellref.all:
            for mec in mechanisms:
                if mec in dir(sec(0.5)):
                    sec.uninsert(mec)


def _update_gpas(node_manager, filename, gjc, correction_iteration_load):
    """Update the g_pas attribute of certain cells
    Cell ids, segment names and new g_pas values are provided in the input file
    """
    import h5py

    processed_cells = 0
    try:
        g_pas_file = h5py.File(filename, "r")
    except OSError as e:
        raise ConfigurationError(f"Error opening g_pas file {filename}") from e
    raw_cell_gids = node_manager.local_nodes.gids(raw_gids=True)
    offset = node_manager.local_nodes.offset
    if f"g_pas/{gjc}" not in g_pas_file:
        logging.warning("Data for g_pas/%s not found in %s", gjc, filename)
        return 0
    for agid in g_pas_file[f"g_pas/{gjc}/"]:
        gid = int(agid[1:])
        if gid in raw_cell_gids:  # if the node has a part of the cell
            cell = node_manager.getCell(gid + offset)
            processed_cells += 1
            for sec in cell.all:
                for seg in sec:
                    try:
                        attr_name = str(seg)[str(seg).index(".") + 1 :]
                        value = g_pas_file[f"g_pas/{gjc}/{agid}"][attr_name][
                            correction_iteration_load
                        ]
                    except Exception as e:
                        raise ConfigurationError(
                            f"Failed to load data in g_pas file {filename}: {e}"
                        ) from e
                    seg.g_pas = value
    return processed_cells


def _load_holding_ic(node_manager, filename, gjc):
    """Add IClamps to certain cells for holding the membrane voltage
    Cell ids, current amplitudes are provided in the input file
    """
    import h5py

    holding_ic_per_gid = {}
    try:
        holding_per_gid = h5py.File(filename, "r")
    except OSError as e:
        raise ConfigurationError(f"Error opening MEComboInfo file {filename}") from e

    if f"holding_per_gid/{gjc}" not in holding_per_gid:
        logging.warning("Data for holding_per_gid/%s not found in %s", gjc, holding_per_gid)
        return holding_ic_per_gid
    raw_cell_gids = node_manager.local_nodes.gids(raw_gids=True)
    offset = node_manager.local_nodes.offset
    for agid in holding_per_gid["holding_per_gid"][str(gjc)]:
        gid = int(agid[1:])
        if gid in raw_cell_gids:
            final_gid = gid + offset
            holding_ic_per_gid[final_gid] = Nd.h.IClamp(
                0.5, sec=node_manager.getCell(final_gid).soma[0]
            )
            holding_ic_per_gid[final_gid].dur = 9e9
            try:
                holding_ic_per_gid[final_gid].amp = holding_per_gid["holding_per_gid"][str(gjc)][agid][()]  # noqa: E501 #fmt: skip
            except Exception as e:
                raise ConfigurationError(
                    f"Failed to load data in g_pas file {filename}: {e}"
                ) from e
    return holding_ic_per_gid


def _find_holding_current(node_manager, filename):
    """Add SEClamps to certain cells for holding the membrane voltage
    Cell ids, voltage amplitude are provided in the input file
    """
    import h5py

    try:
        v_per_gid = h5py.File(filename, "r")
    except OSError as e:
        raise ConfigurationError(f"Error opening voltage file {filename}") from e

    logging.info("Inject voltage clamps without disabling holding current!")

    seclamp_per_gid = {}
    raw_cell_gids = node_manager.local_nodes.gids(raw_gids=True)
    offset = node_manager.local_nodes.offset
    for agid in v_per_gid["v_per_gid"]:
        gid = int(agid[1:])
        if gid in raw_cell_gids:
            final_gid = gid + offset
            seclamp_per_gid[final_gid] = Nd.h.SEClamp(
                0.5, sec=node_manager.getCell(final_gid).soma[0]
            )
            seclamp_per_gid[final_gid].dur1 = 9e9
            seclamp_per_gid[final_gid].amp1 = float(v_per_gid["v_per_gid"][agid][()])
            seclamp_per_gid[final_gid].rs = 0.0000001
    return seclamp_per_gid



"""Main module for handling and instantiating synaptical connections"""

import logging
import pickle  # noqa: S403
from pathlib import Path

import numpy as np

from .connection_manager import ConnectionManagerBase
from .core import MPI, NeuronWrapper as Nd
from .core.configuration import ConfigurationError, SimConfig
from .gap_junction_user_corrections import load_user_modifications
from .io.sonata_config import ConnectionTypes
from .io.synapse_reader import SonataReader, SynapseParameters


class GapJunctionConnParameters(SynapseParameters):
    """Glial-glial gap junction connection parameters.

    This class overrides the `_fields` attribute from the base `SynapseParameters`
    class to define a specific set of parameters relevant for gap junctions.

    The `_optional` and `_reserved` dictionaries are inherited from the base class
    and apply unchanged.

    Note:
        - Only the `_fields` dictionary is overridden.
        - The dtype construction and utility methods are reused as-is.
    """

    # Attribute names of synapse parameters, consistent with the normal synapses
    _fields = {
        "sgid": np.int64,
        "isec": np.int64,
        "offset": np.float64,
        "weight": np.float64,
        "efferent_junction_id": np.int64,
        "afferent_junction_id": np.int64,
        "ipt": np.float64,
        "location": np.float64,
    }


class GapJunctionSynapseReader(SonataReader):
    Parameters = GapJunctionConnParameters
    parameter_mapping = {
        "weight": "conductance",
    }
    # "isec", "ipt", "offset" are custom parameters as in base class


class GapJunctionManager(ConnectionManagerBase):
    """The GapJunctionManager is similar to the SynapseRuleManager. It will
    open dedicated connectivity files which will have the locations and
    conductance strengths of gap junctions detected in the circuit.
    The user will have the capacity to scale the conductance weights.
    """

    CONNECTIONS_TYPE = ConnectionTypes.GapJunction
    SynapseReader = GapJunctionSynapseReader

    def __init__(self, gj_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        """Initialize GapJunctionManager, opening the specified GJ
        connectivity file.

        Args:
            gj_conf: The gaps junctions configuration block / dict
            target_manager: The TargetManager which will be used to query
                targets and translate locations to points
            cell_manager: The cell manager of the target population
            src_cell_manager: The cell manager of the source population
        """
        if cell_manager.circuit_target is None:
            raise ConfigurationError(
                "No circuit target. Required when initializing GapJunctionManager"
            )
        if "Path" not in gj_conf:
            raise ConfigurationError("Missing GapJunction 'Path' configuration")

        super().__init__(gj_conf, target_manager, cell_manager, src_cell_manager, **kw)
        self._src_target_filter = target_manager.get_target(
            cell_manager.circuit_target, src_cell_manager.population_name
        )
        self.holding_ic_per_gid = None
        self.seclamp_per_gid = None
        self.seclamp_current_per_gid_recorder = None

    def create_connections(self, *_, **_kw):
        """Gap Junctions dont use connection blocks, connect all belonging to target"""
        self.connect_all()

    def configure_connections(self, conn_conf):
        """Gap Junctions dont configure_connections"""

    def finalize(self, *_, **_kw):
        super().finalize(conn_type="Gap-Junctions")
        if (
            gj_target_pop := SimConfig.beta_features.get("gapjunction_target_population")
        ) and self.cell_manager.population_name == gj_target_pop:
            logging.info("Load user modification on %s", self)
            self.holding_ic_per_gid, self.seclamp_per_gid = load_user_modifications(self)
            if self.seclamp_per_gid:
                # Record seclamp currents for saving to a file at the end
                self.seclamp_current_per_gid_recorder = {}
                for gid, seclamp in self.seclamp_per_gid.items():
                    self.seclamp_current_per_gid_recorder[gid] = Nd.h.Vector()
                    self.seclamp_current_per_gid_recorder[gid].record(seclamp._ref_i)

    @staticmethod
    def _finalize_conns(_final_tgid, conns, *_, **_kw):
        for conn in reversed(conns):
            conn.finalize_gap_junctions()
        return len(conns)

    def save_seclamp(self):
        """Save seclamps to a file"""
        if self.seclamp_current_per_gid_recorder:
            logging.info("Save SEClamp currents for gap junction user corrections")
            vals = {
                gid: hoc_vec.as_numpy()
                for gid, hoc_vec in self.seclamp_current_per_gid_recorder.items()
            }
            output_dir = Path(SimConfig.output_root) / "gap_junction_seclamps"
            output_dir.mkdir(parents=True, exist_ok=True)
            with open(output_dir / f"data_for_host_{MPI.rank}.p", "wb") as f:
                pickle.dump(vals, f)



"""Definitions for implementing new Engines to handle different cell types"""

import importlib
import logging
import os
import pkgutil


class _EngineMeta(type):
    """A metaclass providing registration for new Engines"""

    __engines = {}
    __connection_types = {}

    def __init__(cls, name, bases, attrs):
        type.__init__(cls, name, bases, attrs)
        ename = name.replace("Engine", "")
        logging.info(" * Registering Engine %s (%s.%s)", ename, cls.__module__, name)
        cls.__engines.setdefault(ename, cls)
        cls.__connection_types.update(cls.ConnectionTypes)

    @property
    def engines(cls):
        return cls.__engines.values()

    @property
    def connection_types(cls):
        return cls.__connection_types

    def get(cls, name):
        """Each engine is a singleton"""
        if name is None:
            return True  # Not setting means use default
        if name in cls.__engines:
            return cls.__engines[name]
        raise RuntimeError("Engine could not be found: " + name)

    def find_plugins(cls):
        # Find/register plugins
        plugin_module = os.environ.get("NEURODAMUS_PLUGIN")
        if plugin_module:
            importlib.import_module(plugin_module)
        # Auto import correctly named modules
        for _finder, name, _ispkg in pkgutil.iter_modules():
            if name.startswith("neurodamus_"):
                importlib.import_module(name)
        logging.info(" => Engines Available: %s", list(cls.__engines.keys()))
        logging.info("  : Connections Types: %s", list(cls.__connection_types.keys()))


class EngineBase(metaclass=_EngineMeta):
    """A base class to define an engine.

    Engines are the fundamental blocks to handle different kinds of cells, like
    Glia, in a plugin-like interface.
    Engines must either implement their own create_cells and create_synapses
    methods (for full flexibility) or specify which are the Manager classes.

    Specifying the Manager classes is suitable for most cases.
    Without any customization an engine will default to use:

      CellManagerCls = None   # Use existing instance of CellDistributor
      InnerConnectivityCls = None  # Use existing instance of SynapseRuleManager

    Such setup is equivalent to not specifying the Engine for a given circuit,
    effectively handling additional circuits by the built-in Engine.
    Specifying CellManagerCls will instantiate cells with the new Engine. If a
    InnerConnectivityCls is not provided then only cell creation happens.
    Specifying InnerConnectivityCls alone is not supported.

    """

    CellManagerCls = None
    InnerConnectivityCls = None
    ConnectionTypes = {}
    """A dict of the new connection types and associated Manager class"""
    CircuitPrecedence = 1
    """Precedence influences instantiation order. The lower the earlier setup"""



"""Mechanisms to load and balance cells across the computing resources."""

import abc
import hashlib
import logging  # active only in rank 0 (init)
import os
import weakref
from contextlib import contextmanager
from io import StringIO
from pathlib import Path

import numpy as np

from .connection_manager import ConnectionManagerBase
from .core import (
    MPI,
    MComplexLoadBalancer,
    NeuronWrapper as Nd,
    ProgressBarRank0 as ProgressBar,
    mpi_no_errors,
    run_only_rank0,
)
from .core.configuration import (
    ConfigurationError,
    GlobalConfig,
    LoadBalanceMode,
    LogLevel,
    SimConfig,
)
from .core.nodeset import SelectionNodeSet
from .io import cell_readers
from .lfp_manager import LFPManager
from .metype import Cell_V6, EmptyCell, PointCell
from .target_manager import TargetSpec
from .utils import compat
from .utils.logging import log_all, log_verbose
from .utils.memory import DryRunStats, get_mem_usage_kb


class VirtualCellPopulation:
    """A virtual cell population offers a compatible interface with Cell Manager,
    however it doesnt instantiate cells.
    It is mostly used as source of projections
    """

    def __init__(self, population_name, gids=None, circuit_target=None):
        """Initializes a VirtualCellPopulation

        A virtual manager will have minimal set of attributes, namely
        the population name, the target name and the SelectionNodeSet
        """
        self.population_name = population_name
        self.circuit_target = circuit_target
        self.local_nodes = SelectionNodeSet(gids).register_global(population_name)

    is_virtual = property(lambda _self: True)

    def __str__(self):
        return f"([VIRT] {self.population_name:s})"


class _CellManager(abc.ABC):
    @abc.abstractmethod
    def get_final_gids(self): ...

    @abc.abstractmethod
    def get_cell(self, gid): ...

    def get_cellref(self, gid):
        """Retrieve a cell object given its gid.
        Note that this function handles multisplit cases incl converting to an
        spgid automatically \n
        Returns: Cell object
        """
        if self._binfo:
            # are we in load balance mode? raw gids are in the binfo
            gid_offset = self._local_nodes.offset
            gid = self._binfo.thishost_gid(gid - gid_offset) + gid_offset
        return self._pc.gid2obj(gid)

    # Methods for compat with hoc

    @abc.abstractmethod
    def getGidListForProcessor(self): ...

    def getCell(self, gid):
        return self.get_cellref(gid)


class CellManagerBase(_CellManager):
    CellType = NotImplemented  # please override
    """The underlying Cell type class

    signature:
        __init__(self, gid, cell_info, circuit_conf)
    """

    _node_loader = None
    """Default function implementing the loading of nodes data

    signature:
        load(circuit_conf, gidvec, stride=1, stride_offset=0)
    """

    def __init__(self, circuit_conf, target_manager, _run_conf=None, **_kw):
        """Initializes CellDistributor

        Args:
            circuit_conf: The "Circuit" config block
            target_parser: the target parser hoc object, to retrieve target cells' info
            run_conf: Run configuration

        """
        self._circuit_conf = circuit_conf
        self._target_manager = target_manager
        self._target_spec = TargetSpec(circuit_conf.CircuitTarget)
        self._population_name = None
        self._local_nodes = None
        self._total_cells = 0  # total cells in target, being simulated
        self._gid2cell = {}

        self._ionchannel_seed = 0
        self._binfo = None
        self._pc = Nd.pc
        self._conn_managers_per_src_pop = weakref.WeakValueDictionary()

        if isinstance(circuit_conf.CellLibraryFile, str):
            self._init_config(circuit_conf, self._target_spec.population)
        else:
            logging.info(" => %s Circuit has been disabled", self.circuit_name or "(default)")

    # read-only properties
    target_manager = property(lambda self: self._target_manager)
    local_nodes = property(
        lambda self: self._local_nodes if self._local_nodes is not None else SelectionNodeSet()
    )
    total_cells = property(lambda self: self._total_cells)
    cells = property(lambda self: self._gid2cell.values())
    gid2cell = property(lambda self: self._gid2cell)
    pc = property(lambda self: self._pc)
    population_name = property(lambda self: self._population_name)
    circuit_target = property(lambda self: self._target_spec.name)
    circuit_name = property(lambda self: self._circuit_conf.name)
    is_virtual = property(lambda _self: False)
    connection_managers = property(lambda self: self._conn_managers_per_src_pop)

    def is_initialized(self):
        return self._local_nodes is not None

    def __str__(self):
        return f"({self.__class__.__name__}: {self._population_name!s})"

    # Compatibility with neurodamus-core (used by TargetManager, CompMapping)
    # Create hoc vector from numpy.array
    def getGidListForProcessor(self):
        return compat.hoc_vector(self.local_nodes.gids(raw_gids=False))

    def get_final_gids(self):
        return np.array(self.local_nodes.gids(raw_gids=False))

    def _init_config(self, circuit_conf, pop):
        if not pop:  # Last attempt to get pop name
            pop = self._get_sonata_population_name(circuit_conf.CellLibraryFile)
            logging.info(" -> Discovered node population name: %s", pop)
        self._population_name = pop
        self._local_nodes = SelectionNodeSet().register_global(pop)

    @staticmethod
    def _get_sonata_population_name(node_file):
        import libsonata  # only for SONATA

        pop_names = libsonata.NodeStorage(node_file).population_names
        if len(pop_names) != 1:
            raise ConfigurationError(
                "Could not determine population name from sonata file circuit."
            )
        return next(iter(pop_names), None)

    def load_nodes(self, load_balancer=None, *, _loader=None, loader_opts=None):
        """Top-level loader of nodes."""
        if self._local_nodes is None:
            return
        conf = self._circuit_conf
        _loader = _loader or self._node_loader
        cycle = loader_opts.pop("cycle_i", None)
        loader_f = (lambda *args: _loader(*args, **loader_opts)) if loader_opts else _loader

        logging.info("Reading Nodes (METype) info from '%s'", conf.CellLibraryFile)
        if (
            load_balancer
            and hasattr(load_balancer, "population")
            and load_balancer.population != self._target_spec.population
        ):
            log_verbose("Load balance object doesn't apply to '%s'", self._target_spec.population)
            load_balancer = None
        if not load_balancer or SimConfig.dry_run:
            gidvec, me_infos, *cell_counts = self._load_nodes(loader_f)
        elif load_balancer and SimConfig.loadbal_mode == LoadBalanceMode.Memory:
            gidvec, me_infos, *cell_counts = self._load_nodes_balance_mem(
                loader_f, load_balancer, cycle
            )
        else:
            gidvec, me_infos, *cell_counts = self._load_nodes_balance(loader_f, load_balancer)
        self._local_nodes.add_gids(gidvec, me_infos)
        self._total_cells = cell_counts[0]
        logging.info(" => Loaded info about %d target cells (out of %d)", *cell_counts)

    def _load_nodes(self, loader_f):
        """Base loader which handles targets"""
        conf = self._circuit_conf
        target_spec = self._target_spec

        if target_spec.name:
            logging.info(" -> Distributing '%s' target cells Round-Robin", target_spec)
            target_gids = self._target_manager.get_target(target_spec).gids(raw_gids=True)
            gidvec, me_infos, full_size = loader_f(conf, target_gids, MPI.size, MPI.rank)
            total_cells = len(target_gids)
        else:
            logging.info(" -> Distributing ALL cells Round-Robin (No target provided)")
            gidvec, me_infos, full_size = loader_f(conf, None, MPI.size, MPI.rank)
            total_cells = full_size
        return gidvec, me_infos, total_cells, full_size

    def _load_nodes_balance(self, loader_f, load_balancer):
        target_spec = self._target_spec
        if not load_balancer.valid_load_distribution(target_spec):
            raise RuntimeError(
                "No valid Load Balance info could be found or derived."
                "Please perform a full load balance."
            )

        logging.info(" -> Distributing target '%s' using Load-Balance", target_spec.name)
        self._binfo = load_balancer.load_balance_info(target_spec)
        # self._binfo has gidlist, but gids can appear multiple times
        all_gids = np.unique(self._binfo.gids.as_numpy().astype("uint32"))
        total_cells = len(all_gids)
        gidvec, me_infos, full_size = loader_f(self._circuit_conf, all_gids)
        return gidvec, me_infos, total_cells, full_size

    def _load_nodes_balance_mem(self, loader_f, load_balancer, cycle_i):
        targetspec: TargetSpec = self._target_spec

        population = targetspec.population
        if population in load_balancer:
            all_gids = load_balancer.get(population).get((MPI.rank, cycle_i), [])
            all_gids = np.array(all_gids, dtype="uint32")
            total_cells = len(all_gids)
            logging.debug("Loading %d cells in rank %d", total_cells, MPI.rank)
            if total_cells == 0:
                return [], [], 0, 0
            gidvec, me_infos, full_size = loader_f(self._circuit_conf, all_gids)
            return gidvec, me_infos, total_cells, full_size
        return self._load_nodes(loader_f)

    # -
    def finalize(self, **opts):
        """Instantiates cells and initializes the network in the simulator.

        Note: it should be called after all cell distributors have done load_nodes()
            so gids offsets are final.
        """
        if self._local_nodes is None:
            return
        logging.info("Finalizing cells... Gid offset: %d", self._local_nodes.offset)
        self._instantiate_cells(**opts)
        self._update_targets_local_gids()
        self._init_cell_network()
        self._local_nodes.clear_cell_info()

    @mpi_no_errors
    def _instantiate_cells(self, cell_type=None, **_opts):
        cell_type = cell_type or self.CellType
        if SimConfig.crash_test_mode:
            cell_type = PointCell

        assert cell_type is not None, "Undefined cell_type in Manager"
        Nd.execute("xopen_broadcast_ = 0")

        logging.info(" > Instantiating cells... (%d in Rank 0)", len(self._local_nodes))
        cell_offset = self._local_nodes.offset

        if GlobalConfig.verbosity >= LogLevel.DEBUG:
            gid_info_iter = self._local_nodes.iter_cell_info()
        else:
            gid_info_iter = ProgressBar.iter(
                self._local_nodes.iter_cell_info(), len(self._local_nodes)
            )

        for gid, cell_info in gid_info_iter:
            cell = cell_type(gid, cell_info, self._circuit_conf)
            self._store_cell(gid + cell_offset, cell)

    @mpi_no_errors
    def _instantiate_cells_dry(self, cell_type, skip_metypes, **_opts):
        """Instantiates the subset of selected cells while measuring memory taken by each metype

        Args:
            cell_type: The cell type class
            full_memory_counter: The memory counter to be updated for each metype
        """
        assert cell_type is not None, "Undefined cell_type in Manager"
        Nd.execute("xopen_broadcast_ = 0")

        logging.info(" > Dry run on cells... (%d in Rank 0)", len(self._local_nodes))
        cell_offset = self._local_nodes.offset

        prev_metype = None
        prev_memory = get_mem_usage_kb()
        metype_n_cells = 0
        memory_dict = {}
        MAX_CELLS = 50

        def store_metype_stats(metype, n_cells):
            nonlocal prev_memory
            end_memory = get_mem_usage_kb()
            memory_allocated = end_memory - prev_memory
            log_all(
                logging.DEBUG,
                " * METype %s: %.1f KiB averaged over %d cells",
                metype,
                memory_allocated / n_cells,
                n_cells,
            )
            memory_dict[metype] = max(0, memory_allocated / n_cells)
            prev_memory = end_memory

        for gid, cell_info in self._local_nodes.iter_cell_info():
            if cell_info is None:
                continue
            metype = f"{cell_info.mtype}-{cell_info.etype}"
            if metype in skip_metypes:
                continue
            if prev_metype is not None and metype != prev_metype:
                store_metype_stats(prev_metype, metype_n_cells)
                metype_n_cells = 0
            if metype_n_cells >= MAX_CELLS:
                continue
            cell = cell_type(gid, cell_info, self._circuit_conf)
            self._store_cell(gid + cell_offset, cell)
            prev_metype = metype
            metype_n_cells += 1

        if prev_metype is not None and metype_n_cells > 0:
            store_metype_stats(prev_metype, metype_n_cells)

        return memory_dict

    def _update_targets_local_gids(self):
        logging.info(" > Updating targets")
        # Add local gids to matching targets
        self._target_manager.register_local_nodes(self._local_nodes)

    def _init_cell_network(self):
        """Init global gids for cell networking"""
        logging.info(" > Initializing cell network")
        self._init_rng()
        pc = self._pc

        for final_gid, cell in self._gid2cell.items():
            cell.re_init_rng(self._ionchannel_seed)
            nc = cell.connect2target(None)  # Netcon doesnt require being stored
            raw_gid = final_gid - self._local_nodes.offset
            if self._binfo:
                gid_i = int(self._binfo.gids.indwhere("==", raw_gid))
                cb = self._binfo.bilist.object(self._binfo.cbindex.x[gid_i])
                # multisplit cells call cb.multisplit() instead
                if cb.subtrees.count() > 0:
                    cb.multisplit(nc, self._binfo.msgid, pc, pc.id())
                    cell.gid = final_gid
                    cell.raw_gid = raw_gid
                    continue

            pc.set_gid2node(final_gid, pc.id())
            pc.cell(final_gid, nc)
            cell.gid = final_gid  # update the cell.gid last (RNGs had to use the base gid)
            cell.raw_gid = raw_gid

        pc.multisplit()

    def load_artificial_cell(self, gid, artificial_cell):
        logging.info(" > Adding Artificial cell for CoreNeuron")
        cell = EmptyCell(gid, artificial_cell)
        self._store_cell(gid, cell)
        nc = cell.connect2target(None)
        self._pc.set_gid2node(cell.gid, self._pc.id())
        self._pc.cell(cell.gid, nc)

    def _init_rng(self):
        rng_info = Nd.RNGSettings()
        self._ionchannel_seed = rng_info.getIonChannelSeed()
        return rng_info

    def _store_cell(self, gid, cell):
        self._gid2cell[gid] = cell

    def get_cell(self, gid):
        return self._gid2cell[gid]

    def get_cellref(self, gid):
        return self._gid2cell[gid]._cellref

    def record_spikes(self, gids=None, append_spike_vecs=None):
        """Setup recording of spike events (crossing of threshold) for cells on this node"""
        if not self._local_nodes:
            return None
        spikevec, idvec = append_spike_vecs or (Nd.Vector(), Nd.Vector())
        if gids is None:
            gids = self._local_nodes.gids(raw_gids=False)
            gid_offset = self._local_nodes.offset

        for gid in gids:
            # only want to collect spikes of cell pieces with the soma (i.e. the real gid)
            if not self._binfo or self._binfo.thishost_gid(gid - gid_offset) + gid_offset == gid:
                self._pc.spike_record(gid, spikevec, idvec)
        return spikevec, idvec

    def register_connection_manager(self, conn_manager: ConnectionManagerBase):
        src_population = conn_manager.src_node_population
        if src_population in self._conn_managers_per_src_pop:
            logging.warning("Skip registering %s as a second pop source", conn_manager)
        else:
            self._conn_managers_per_src_pop[src_population] = conn_manager

    def pre_stdinit(self):
        """Post stdinit actions"""
        if self._circuit_conf.DetailedAxon:
            log_verbose("Now deleting the axon! 🤩")
            for c in self.cells:
                c.delete_axon()


class GlobalCellManager(_CellManager):
    """GlobalCellManager is a wrapper over all Cell Managers so that we can query
    any cell from its global gid
    """

    def __init__(self):
        self._cell_managers = []
        self._pc = Nd.pc
        self._lfp_manager = LFPManager()

    def register_manager(self, cell_manager):
        self._cell_managers.append(cell_manager)

    def finalize(self):
        self._cell_managers.sort(key=lambda x: x.local_nodes.offset)

    # Accessor methods (Keep CamelCase API for compatibility with existing hoc)
    # ----------------
    def getGidListForProcessor(self):
        def _hoc_append(vec_a, vec_b):
            return vec_a.append(vec_b)

        from functools import reduce

        return reduce(_hoc_append, (man.getGidListForProcessor() for man in self._cell_managers))

    def get_final_gids(self):
        return np.concatenate([man.get_final_gids() for man in self._cell_managers])

    def _find_manager(self, gid):
        cell_managers_iter = iter(self._cell_managers)
        prev_manager = next(cell_managers_iter)  # base cell manager
        for manager in cell_managers_iter:
            if manager.local_nodes.offset > gid:
                break
            prev_manager = manager
        return prev_manager

    def get_cell(self, gid):
        manager = self._find_manager(gid)
        return manager.get_cell(gid)

    def get_cellref(self, gid):
        """Retrieve a cell object given its gid.
        Note that this function handles multisplit cases incl converting to an
        spgid automatically \n
        Returns: Cell object
        """
        manager = self._find_manager(gid)
        if manager._binfo:
            # are we in load balance mode? raw gids are in the binfo
            gid_offset = manager.local_nodes.offset
            gid = manager._binfo.thishost_gid(gid - gid_offset) + gid_offset
        return self._pc.gid2obj(gid)

    def getSpGid(self, gid):
        """Retrieve the spgid from a gid (provided we are using loadbalancing)

        Args:
            gid: The base gid (as read from start.ncs)

        Returns: The gid as it appears on this cpu (if this is the same as the base gid,
        then that is the soma piece)
        """
        manager = self._find_manager(gid)
        if manager._binfo:
            gid_offset = manager.local_nodes.offset
            return manager._binfo.thishost_gid(gid - gid_offset) + gid_offset
        return gid

    def getPopulationInfo(self, gid):
        manager = self._find_manager(gid)
        return manager.population_name, manager.local_nodes.offset


class CellDistributor(CellManagerBase):
    """Manages a group of cells for BBP simulations, V5 and V6

    Instantiated cells are stored locally (.cells property)
    """

    _sonata_with_extra_attrs = True  # Enable search extra node attributes

    def _init_config(self, circuit_conf, _pop):
        if not circuit_conf.CellLibraryFile:
            raise ConfigurationError("CellLibraryFile not set")

        super()._init_config(circuit_conf, _pop)

    def load_nodes(self, load_balancer=None, **kw):
        """Gets gids from target, splits and returns a GidSet with all metadata"""
        loader_opts = kw.pop("loader_opts", {}).copy()
        all_cell_requirements = SimConfig.cell_requirements
        cell_requirements = all_cell_requirements.get(self._population_name)
        loader = cell_readers.load_sonata
        loader_opts["node_population"] = self._population_name  # mandatory in Sonata
        loader_opts["load_dynamic_props"] = cell_requirements
        loader_opts["has_extra_data"] = self._sonata_with_extra_attrs

        log_verbose("Nodes Format: SONATA , Loader: %s", loader.__name__)
        return super().load_nodes(load_balancer, _loader=loader, loader_opts=loader_opts)

    def _instantiate_cells(self, dry_run_stats_obj: DryRunStats = None, **opts):
        """Instantiates cells, honouring dry_run if provided

        Args:
            dry_run_store_stats: Do a dry-run and update the inner fields accordingly
        """
        if self.CellType is not NotImplemented:
            return super()._instantiate_cells(self.CellType)

        conf = self._circuit_conf
        cell_type = Cell_V6
        if conf.MorphologyType:
            cell_type.morpho_extension = conf.MorphologyType

        log_verbose("Loading metypes from: %s", conf.METypePath)
        log_verbose(
            "Loading '%s' morphologies from: %s", cell_type.morpho_extension, conf.MorphologyPath
        )
        if dry_run_stats_obj is None:
            super()._instantiate_cells(cell_type, **opts)
        else:
            cur_metypes_mem = dry_run_stats_obj.metype_memory
            memory_dict = self._instantiate_cells_dry(cell_type, cur_metypes_mem, **opts)
            log_verbose("Updating global dry-run memory counters with %d items", len(memory_dict))
            cur_metypes_mem.update(memory_dict)

        return None


class LoadBalance:
    """Class handling the several types of load_balance info, including
    generating and loading the various files.

    LoadBalance instances target the current system (cpu count) and circuit
    BUT check/create load distribution for any given target.
    The circuit is identified by the nodes file AND population.

    NOTE: Given the heavy costs of computing load balance, some state files are created
    which allow the balance info to be reused. These are

     - cx_{TARGET}.dat: File with complexity information for the cells of a given target
     - cx_{TARGET}.{CPU_COUNT}.dat: The file assigning cells/pieces to individual CPUs ranks.

    For more information refer to the developer documentation.

    :param balance_mode: Mode of balancing the load, as defined by
        :class:`neurodamus.core.configuration.LoadBalanceMode`.
    :param nodes_path: Path to the nodes file.
    :param pop: Population identifier for the circuit.
    :param target_manager: Manages target information for load balancing.
    :param target_cpu_count: Optional number of CPUs for the target.
    """

    _base_output_dir = "sim_conf"
    _circuit_lb_dir_tpl = "_loadbal_%s.%s"  # Placeholders are (file_src_hash, population)
    _cx_filename_tpl = "cx_%s#.dat"  # use # to well delimiter the target name
    _cpu_assign_filename_tpl = "cx_%s#.%s.dat"  # prefix must be same (imposed by Neuron)

    def __init__(self, balance_mode, nodes_path, pop, target_manager, target_cpu_count=None):
        """Creates a new Load Balance object, associated with a given node file"""
        self.lb_mode = balance_mode
        self.target_cpu_count = target_cpu_count or MPI.size
        self._target_manager = target_manager
        self._valid_loadbalance = set()
        self.population = pop or ""
        self._lb_dir, self._cx_targets = self._get_circuit_loadbal_dir(nodes_path, self.population)
        log_verbose("Found existing targets with loadbal: %s", self._cx_targets)

    @classmethod
    @run_only_rank0
    def _get_circuit_loadbal_dir(cls, node_file, pop) -> tuple:
        """Ensure lbal dir exists. dir may be crated on rank 0"""
        lb_dir = cls._loadbal_dir(node_file, pop)
        if lb_dir.is_dir():
            return lb_dir, cls._get_lbdir_targets(lb_dir)

        logging.info("Creating load balance dir %s", lb_dir)
        os.makedirs(lb_dir)
        return lb_dir, set()

    @classmethod
    def _get_lbdir_targets(cls, lb_dir: Path) -> list:
        """Inspects the load-balance folder and detects which targets are load balanced"""
        prefix, suffix = cls._cx_filename_tpl.split("%s")
        return {
            fname.name[len(prefix) : -len(suffix)]
            for fname in lb_dir.glob(cls._cx_filename_tpl.replace("%s", "*"))
        }

    @run_only_rank0
    def valid_load_distribution(self, target_spec: TargetSpec) -> bool:
        """Checks whether we have valid load-balance files, attempting to
        derive from larger target distributions if possible.
        """
        if (target_spec.population) != self.population:
            logging.info(
                " => Load balance Population mismatch. Requested: %s, Existing: %s",
                target_spec.population,
                self.population,
            )
            return False

        target_name = target_spec.simple_name

        # Check cache
        if target_name in self._valid_loadbalance:
            logging.info(" => Cell distribution from Load Balance is valid")
            return True

        # A valid load-distribution requires BOTH the cx and the cpu_assign files.
        if self._cx_valid(target_spec):
            cpu_assign_filename = self._cpu_assign_filename(target_name)
            if cpu_assign_filename.is_file():
                logging.info(" => Found valid load balance: %s", cpu_assign_filename)
            else:
                # Still need to derive the CPU assignment
                logging.info(" => Found valid complexity file: %s", self._cx_filename(target_name))
                self._cpu_assign(target_name)
            self._valid_loadbalance.add(target_name)
            return True

        # If not found, attempt to find a cx file computed for a superset target
        # so we can derive a new cx file without instantiating cells
        if self._reuse_cell_complexity(target_spec):
            self._cpu_assign(target_name)
            self._valid_loadbalance.add(target_name)
            return True
        return False

    # -
    def _reuse_cell_complexity(self, target_spec: TargetSpec) -> bool:
        """Check if the complexities of all target gids were already calculated
        for another target.
        """
        # Abort if there are no cx files yet or in case now we request full circuit
        # since its impossible to have a superset of it
        if not target_spec.name or not self._cx_targets:
            logging.info(" => Target Cx reusing is not available.")
            return False

        logging.info("Attempt reusing cx files from other targets...")
        target_gids = self._get_target_raw_gids(target_spec)
        cx_other = {}

        for previous_target in self._cx_targets:
            log_verbose("Trying previous cx file on target %s", previous_target)
            ref_filename = self._cx_filename(previous_target)
            if self._cx_contains_gids(ref_filename, target_gids, cx_other):
                break  # done!
            log_verbose("  - Target is not a superset. Ignoring.")
        else:
            logging.info(" => Did not find any suitable target")
            return False

        new_cx_filename = self._cx_filename(target_spec.simple_name)
        logging.info(
            "Target %s is a subset of the target %s. Generating %s",
            target_spec.name,
            previous_target,
            new_cx_filename,
        )

        # Write the new cx file since Neuron needs it to do CPU assignment
        with open(new_cx_filename, "w", encoding="utf-8") as newfile:
            self._write_msdat_dict(newfile, cx_other, target_gids)
        # register
        self._cx_targets.add(target_spec.simple_name)
        return True

    # -
    def _cx_valid(self, target_spec) -> bool:
        """Determine if valid complexity files exist for the provided circuit and
        target spec, otherwise return False -> we need to regen load balance info
        """
        # _cx_targets populated at instantiation. If not there, no chance. If in list, verify
        if not self._cx_targets:
            logging.info(" => No complexity files for current circuit yet")
            return False

        target_name = target_spec.simple_name
        cx_filename = self._cx_filename(target_name)

        if target_name not in self._cx_targets:
            logging.info(" => No Cx files available for requested target")
            return False

        if target_spec:  # target provided, otherwise everything
            target_gids = self._get_target_raw_gids(target_spec)
            if not self._cx_contains_gids(cx_filename, target_gids):
                logging.warning(" => %s invalid: changed target definition!", cx_filename)
                return False
        return True

    @classmethod
    def _cx_contains_gids(cls, cxpath, target_gids, out_cx=None) -> bool:
        """Checks a cx file contains complexities for given gids"""
        if not cxpath.is_file():
            log_verbose("  - cxpath doesnt exist: %s", cxpath)
            return False
        with open(cxpath, encoding="utf-8") as f:
            cx_saved = cls._read_msdat(f)
        if not set(cx_saved.keys()) >= set(target_gids):
            log_verbose("  - Not all GIDs in target %s %s", set(cx_saved.keys()), set(target_gids))
            return False
        if out_cx is not None:
            out_cx.update(cx_saved)
        return True

    @contextmanager
    def generate_load_balance(self, target_spec, cell_distributor):
        """Context manager that creates load balance for the circuit instantiated within

        Args:
            target_str: a string representation of the target.
            cell_distributor: the cell distributor object to which we can query
                the cells to be load balanced
        """
        mcomplex = MComplexLoadBalancer()  # init mcomplex before building circuit
        yield
        target_str = target_spec.simple_name
        self._compute_save_complexities(target_str, mcomplex, cell_distributor)
        self._cpu_assign(target_str)
        self._valid_loadbalance.add(target_str)

    @mpi_no_errors
    def _compute_save_complexities(self, target_str, mcomplex, cell_distributor):
        msfactor = 1e6 if self.lb_mode == LoadBalanceMode.WholeCell else 0.8
        out_filename = self._cx_filename(target_str)

        cx_cells = self._compute_complexities(mcomplex, cell_distributor)
        total_cx, max_cx = self._cell_complexity_total_max(cx_cells)
        lcx = self._get_optimal_piece_complexity(total_cx, self.target_cpu_count, msfactor)
        logging.info(
            "LB Info: TC=%.3f MC=%.3f OptimalCx=%.3f FileName=%s",
            total_cx,
            max_cx,
            lcx,
            out_filename,
        )

        ms_list = []
        tmp = Nd.Vector()

        for cell in cell_distributor.cells:
            mcomplex.cell_complexity(cell.CellRef)
            mcomplex.multisplit(cell.raw_gid, lcx, tmp)
            ms_list.append(tmp.c())

        # To output build independently the contents of the file then append
        ostring = StringIO()
        for ms in ms_list:
            self._write_msdat(ostring, ms)

        all_ranks_cx = MPI.py_gather(ostring.getvalue(), 0)
        if MPI.rank == 0:
            with open(out_filename, "w", encoding="utf-8") as fp:
                fp.write(f"1\n{cell_distributor.total_cells}\n")
                fp.writelines(all_ranks_cx)

        # register
        self._cx_targets.add(target_str)

    @staticmethod
    def _cell_complexity_total_max(cx_cells):
        """Returns: Tuple of (TotalComplexity, max_complexity)"""
        local_max = max(cx_cells) if len(cx_cells) > 0 else 0.0
        local_sum = sum(cx_cells) if len(cx_cells) > 0 else 0.0

        global_total = MPI.allreduce(local_sum, MPI.SUM)
        global_max = MPI.allreduce(local_max, MPI.MAX)
        return global_total, global_max

    @classmethod
    def _compute_complexities(cls, mcomplex, cell_distributor):
        cx_cell = compat.Vector("f")
        pc = cell_distributor.pc
        for gid in cell_distributor.local_nodes.gids(raw_gids=False):
            cx_cell.append(mcomplex.cell_complexity(pc.gid2cell(gid)))
        return cx_cell

    @staticmethod
    def _get_optimal_piece_complexity(total_cx, nhost, msfactor):
        """Args:
        total_cx: Total complexity
        nhost: Prospective no of hosts
        """
        lps = total_cx * msfactor / nhost
        return int(lps + 1)

    @run_only_rank0
    def _cpu_assign(self, target_name):
        """Assigns cells to 'prospective_hosts' cpus using mymetis3.
        Results are written to file. basename.<NCPU>.dat
        """
        logging.info("Assigning Cells <-> %d CPUs [mymetis3]", self.target_cpu_count)
        base_filename = self._cx_filename(target_name, basename_str=True)
        Nd.mymetis3(base_filename, self.target_cpu_count)

    @staticmethod
    def _write_msdat(fp, ms):
        """Writes load balancing info to an output stream"""
        fp.write(str(int(ms.x[0])))  # gid
        fp.write(f" {ms.x[1]:g}")  # total complexity of cell
        piece_count = int(ms.x[2])
        fp.write(f" {piece_count}\n")
        i = 2

        for _ in range(piece_count):
            i += 1
            subtree_count = int(ms.x[i])
            fp.write(f"  {subtree_count}\n")
            for _ in range(subtree_count):
                i += 1
                cx = ms.x[i]  # subtree complexity
                i += 1
                children_count = int(ms.x[i])
                fp.write(f"   {cx:g} {children_count}\n")
                if children_count > 0:
                    fp.write("    ")
                for _ in range(children_count):
                    i += 1
                    elem_id = int(ms.x[i])  # at next child
                    fp.write(f" {elem_id}")
                if children_count > 0:
                    fp.write("\n")

    @staticmethod
    def _read_msdat(fp):
        """Read load balancing info from an input stream"""
        cx_saved = {}  # dict with key = gid, value = line content
        piece_count = 0
        gid = None
        next(fp)  # first line has 1. skip
        next(fp)  # ncells. skip

        for line in fp:
            if piece_count == 0:
                gid, _cx, piece_count = [int(float(x)) for x in line.split()]
                cx_saved[gid] = [line]
            else:  # Handle parts
                cx_saved[gid].append(line)
                for _ in range(2 * int(line)):  # each subtree has two lines
                    cx_saved[gid].append(next(fp))
                piece_count -= 1

        return cx_saved

    @staticmethod
    def _write_msdat_dict(fp, cx_dict, gids=None):
        """Write out selected gid cx lines from a cx_dict"""
        if gids is None:
            gids = cx_dict.keys()
        fp.write(f"1\n{len(gids)}\n")
        for gid in gids:
            for line in cx_dict[gid]:
                fp.write(line)  # raw lines, include \n

    # -
    def _get_target_raw_gids(self, target_spec) -> np.ndarray:
        return self._target_manager.get_target(target_spec).gids(raw_gids=True)

    def load_balance_info(self, target_spec):
        """Loads a load-balance info for a given target.
        NOTE: Please ensure the load balance exists or is derived before calling this function
        """
        bal_filename = self._cx_filename(target_spec.simple_name, basename_str=True)
        return Nd.BalanceInfo(bal_filename, MPI.rank, MPI.size)

    @classmethod
    def _loadbal_dir(cls, nodefile, population) -> Path:
        """Returns the dir where load balance files are stored for a given nodes file"""
        nodefile_hash = hashlib.md5(nodefile.encode()).hexdigest()[:10]
        return Path(cls._base_output_dir) / (cls._circuit_lb_dir_tpl % (nodefile_hash, population))

    def _cx_filename(self, target_str, basename_str=False) -> Path:
        """Gets the filename of a cell complexity file for a given target"""
        fname = self._lb_dir / (self._cx_filename_tpl % target_str)
        return str(fname)[:-4] if basename_str else fname

    def _cpu_assign_filename(self, target_str) -> Path:
        """Gets the CPU assignment filename for a given target, according to target CPU count"""
        return self._lb_dir / (self._cpu_assign_filename_tpl % (target_str, self.target_cpu_count))

    @staticmethod
    def select_lb_mode(sim_config, run_conf, target):
        """A method which selects the load balance mode according to run config"""
        # Check / set load balance mode
        lb_mode = sim_config.loadbal_mode
        if lb_mode == LoadBalanceMode.MultiSplit:
            if not sim_config.use_coreneuron:
                logging.info("Load Balancing ENABLED. Mode: MultiSplit")
            else:
                logging.warning("Load Balancing mode CHANGED to WholeCell for CoreNeuron")
                lb_mode = LoadBalanceMode.WholeCell

        elif lb_mode == LoadBalanceMode.WholeCell:
            logging.info("Load Balancing ENABLED. Mode: WholeCell")

        elif lb_mode is None:
            if target.is_void():
                lb_mode, reason = LoadBalanceMode.RoundRobin, "No target set, unknown cell count"
            else:
                lb_mode, reason = LoadBalanceMode.auto_select(
                    sim_config.use_neuron, target.gid_count(), run_conf["Duration"]
                )
            logging.warning("Load Balance AUTO-SELECTED: %s. Reason: %s", lb_mode.name, reason)

        return lb_mode



"""Module which defines and handles METypes config (v5/v6 cells)"""

import logging
from abc import abstractmethod
from os import path as ospath

import numpy as np

from .core import NeuronWrapper as Nd
from .core.configuration import ConfigurationError, SimConfig


class SectionIdError(Exception):
    pass


class BaseCell:
    """Class representing an basic cell, e.g. an artificial cell"""

    _section_layout = [
        ("soma", lambda c: c.soma),
        ("axon", lambda c: c.axon, lambda c: int(c.nSecAxonalOrig)),
        ("dend", lambda c: c.dend),
        ("apic", lambda c: c.apic),
        ("ais", lambda c: getattr(c, "ais", [])),
        ("node", lambda c: getattr(c, "node", [])),
        ("myelin", lambda c: getattr(c, "myelin", [])),
    ]

    __slots__ = ("_ccell", "_cellref", "_section_counts", "raw_gid")

    def __init__(self):
        self._cellref = None
        self._ccell = None
        self.raw_gid = None
        self._section_counts = None

    @property
    def CellRef(self):
        return self._cellref

    @property
    def CCell(self):
        return self._ccell

    def re_init_rng(self, ion_seed):
        pass

    def connect2target(self, target_pp=None):
        """Connects empty cell to target"""
        return Nd.NetCon(self._cellref, target_pp)

    def get_section_counts(self):
        """Lazy set of the section counts for the cell."""
        if self._section_counts is None:
            self._section_counts = [
                len(i[1](self._cellref)) if len(i) == 2 else i[2](self._cellref)
                for i in BaseCell._section_layout
            ]
        return self._section_counts

    def get_section_id(self, section):
        """Calculate the global index of a given section within its cell.

        :param cell: The cell instance containing the section of interest
        :param section: The specific section for which the index is required
        :return: The global index of the section, applicable for neuron mapping

        Note: section_id is based on the original cell, before removing the axon.

        Warning: this method returns the original section_id, which may not be valid
        if the axon was removed. The offsets are still calculated based on the
        original cell structure.
        """
        section_name = str(section).rsplit(".", 1)[-1]
        try:
            section_type, index_str = section_name.rsplit("[", maxsplit=1)
            local_idx = int(index_str.rstrip("]"))
            if local_idx < 0:
                raise SectionIdError(f"Negative index {local_idx} in section name: {section_name}")
        except ValueError as e:
            raise SectionIdError(f"Cannot parse section name: {section_name}") from e

        offset = 0
        for name, count in zip(BaseCell._section_layout, self.get_section_counts()):
            name = name[0]
            if name == section_type:
                if local_idx >= count:
                    raise SectionIdError(
                        f"Index {local_idx} out of range for section type '{name}' (count={count})"
                    )
                return offset + local_idx
            offset += count

        raise SectionIdError(f"Unknown section type in: {section_type}")

    def get_sec(self, section_id):
        """Inverse of get_section_id. Given a global section_id, returns the section from the cell.

        :param cell: The cell instance used for offsets
        :param section_id: The global index of the section
        :return: Reference to the section in the cell

        Note: section_id is based on the original cell, before removing the axon.
        Asking for one of the removed sections will raise an error. Asking for one of
        the two remaining sections is still possible. The offsets are still
        calculated based on the original cell structure.
        """
        idx = section_id
        for name, count in zip(BaseCell._section_layout, self.get_section_counts()):
            name, accessor_fn = name[0], name[1]
            if idx < count:
                section_list = accessor_fn(self._cellref)
                if name == "axon" and len(section_list) <= idx:
                    raise SectionIdError(
                        f"The axon was removed ({count} -> {len(section_list)}). "
                        f"The section_id {section_id} refers to a removed axon section "
                        f"(local index {idx})."
                    )
                return section_list[idx]
            idx -= count

        raise SectionIdError(f"Section ID {section_id} is out of bounds.")


class METype(BaseCell):
    """Class representing an METype. Will instantiate a Hoc-level cell as well"""

    morpho_extension = "asc"
    """The extension to be applied to morphology files"""

    KEEP_AXON_FLAG = 400

    __slots__ = (
        "_emodel_name",
        "_hypAmp_current",
        "_netcons",
        "_synapses",
        "_threshold_current",
        "exc_mini_frequency",
        "extra_attrs",
        "inh_mini_frequency",
    )

    def __init__(self, gid, etype_path, emodel, morpho_path, meinfos=None, detailed_axon=False):
        """Instantite a new Cell from METype

        Args:
            gid: Cell gid
            etype_path: path for etypes
            emodel: Emodel name
            morpho_path: path for morphologies
            meinfos: dictionary with v6 infos (if v6 circuit)
        """
        super().__init__()
        self._threshold_current = None
        self._hypAmp_current = None
        self._netcons = []
        self._synapses = None
        self._emodel_name = emodel
        self.exc_mini_frequency = None
        self.inh_mini_frequency = None
        self.extra_attrs = None

        self._instantiate_cell(gid, etype_path, emodel, morpho_path, meinfos, detailed_axon)

    gid = property(
        lambda self: int(self._cellref.gid), lambda self, val: setattr(self._cellref, "gid", val)
    )

    # Ensure no METype instances created. Only Subclasses
    @abstractmethod
    def _instantiate_cell(self, *args):
        """Method which instantiates the cell in the simulator"""

    @property
    def synlist(self):
        return self._synapses

    # Named for compat with still existing HOC modules
    def getThreshold(self):
        return self._threshold_current

    def setThreshold(self, value):
        self._threshold_current = value

    def getHypAmp(self):
        if self._hypAmp_current is None:
            logging.warning("EModel %s doesnt define HypAmp current", self._emodel_name)
            return 0
        return self._hypAmp_current

    def setHypAmp(self, value):
        self._hypAmp_current = value

    def connect2target(self, target_pp=None):
        """Connects MEtype cell to target

        Args:
            target_pp: target point process [default: None]

        Returns: NetCon obj
        """
        if SimConfig.spike_location == "soma":
            sec, seg = self.CellRef.soma[0], self.CellRef.soma[0](1)
        else:
            sec, seg = self.CellRef.axon[1], self.CellRef.axon[1](0.5)
        netcon = Nd.NetCon(seg._ref_v, target_pp, sec=sec)
        netcon.threshold = SimConfig.spike_threshold
        return netcon

    def re_init_rng(self, ion_seed):
        """Re-Init RNG for cell

        Args:
            ion_seed: ion channel seed
        """
        self._ccell.re_init_rng(ion_seed)

    def delete_axon(self):
        pass

    def __del__(self):
        if self._cellref:
            self._cellref.clear()  # cut cyclic reference


class Cell_V6(METype):  # noqa: N801
    __slots__ = ("local_to_global_matrix",)

    def __init__(self, gid, meinfo, circuit_conf):
        mepath = circuit_conf.METypePath
        morpho_path = circuit_conf.MorphologyPath
        detailed_axon = circuit_conf.DetailedAxon
        super().__init__(gid, mepath, meinfo.emodel_tpl, morpho_path, meinfo, detailed_axon)

    def _instantiate_cell(self, gid, etype_path, emodel, morpho_path, meinfos_v6, detailed_axon):
        """Instantiates a SSCx v6 cell"""
        Nd.load_hoc(ospath.join(etype_path, emodel))
        EModel = getattr(Nd, emodel)
        morpho_file = meinfos_v6.morph_name + "." + self.morpho_extension
        keep_axon = detailed_axon and self.KEEP_AXON_FLAG
        add_params = meinfos_v6.add_params or (keep_axon,)  # Keep axon incompatible with add_params

        logging.debug("Loading Gid %d: emodel: %s, Morphology: %s", gid, emodel, morpho_file)
        try:
            # For this step, do not call mpi_abort in neuron and let neurodamus handle and abort,
            # NB: Do not re-raise as ConfigurationError, neurodamus doesn't call mpi_abort so hangs
            old_flag = Nd.pc.mpiabort_on_error(0)
            self._cellref = EModel(gid, morpho_path, morpho_file, *add_params)
            Nd.pc.mpiabort_on_error(old_flag)
        except Exception as e:
            msg = f"Error from NEURON loading Gid {gid}: emodel: {emodel}, morph: {morpho_file}"
            raise RuntimeError(msg) from e
        self._ccell = self._cellref
        self._synapses = Nd.List()
        self._threshold_current = meinfos_v6.threshold_current
        self._hypAmp_current = meinfos_v6.holding_current
        self.exc_mini_frequency = meinfos_v6.exc_mini_frequency
        self.inh_mini_frequency = meinfos_v6.inh_mini_frequency
        self.local_to_global_matrix = meinfos_v6.local_to_global_matrix
        self.extra_attrs = meinfos_v6.extra_attrs

    def local_to_global_coord_mapping(self, points):
        if self.local_to_global_matrix is False:
            raise ConfigurationError(
                "To use local_to_global_coord_mapping please "
                "run neurodamus with `enable_coord_mapping=True`"
            )
        if self.local_to_global_matrix is None:
            raise Exception("Nodes don't provide all 3d position/rotation info")
        return vector_rotate_translate(points, self.local_to_global_matrix)

    def delete_axon(self):
        self._cellref.replace_axon()

    def __getattr__(self, item):
        prop = self.extra_attrs.get(item)
        if prop is None:
            raise AttributeError(item)
        return prop


class EmptyCell(BaseCell):
    """Class representing an empty cell, e.g. an artificial cell
    Workaround for the neuron issue https://github.com/neuronsimulator/nrn/issues/635
    """

    __slots__ = ("gid",)

    def __init__(self, gid, cell):
        super().__init__()
        self._cellref = cell
        self.gid = gid


class PointCell:
    """Class representing a minimal, single section cell for dry-runs"""

    def __init__(self, gid, cell_info, _circuit_conf):
        self.gid = gid
        self.raw_gid = None
        self.soma = [Nd.Section(name="soma[0]", cell=self)]
        self.exc_mini_frequency = float(cell_info.exc_mini_frequency)
        self.inh_mini_frequency = float(cell_info.inh_mini_frequency)
        self._threshold_current = float(cell_info.threshold_current)
        self._hypAmp_current = float(cell_info.holding_current)
        self.synHelperList = []
        self.synlist = []

    CellRef = property(lambda self: self)
    CCell = property(lambda self: self)
    nSecAll = property(lambda _self: 1)  # noqa: N815
    all = property(lambda self: self.soma)
    input_resistance = property(lambda _self: 1)

    def getThreshold(self):
        return self._threshold_current

    def getHypAmp(self):
        return self._hypAmp_current

    def connect2target(self, target_pp=None):
        soma_sec = self.soma[0]
        return Nd.NetCon(soma_sec(1)._ref_v, target_pp, sec=soma_sec)

    def re_init_rng(self, ion_seed):
        pass


# Metadata
# --------


class METypeItem:
    """Metadata about an METype, each possibly used by several cells."""

    __slots__ = (
        "add_params",
        "emodel_tpl",
        "etype",
        "exc_mini_frequency",
        "extra_attrs",
        "holding_current",
        "inh_mini_frequency",
        "local_to_global_matrix",
        "morph_name",
        "mtype",
        "threshold_current",
    )

    def __init__(
        self,
        morph_name,
        etype=None,
        emodel_tpl=None,
        mtype=None,
        threshold_current=0,
        holding_current=0,
        exc_mini_frequency=0,
        inh_mini_frequency=0,
        add_params=None,
        position=None,
        rotation=None,
        scale=1.0,
    ):
        self.morph_name = morph_name
        self.etype = etype
        self.emodel_tpl = emodel_tpl
        self.mtype = mtype
        self.threshold_current = float(threshold_current)
        self.holding_current = float(holding_current)
        self.exc_mini_frequency = float(exc_mini_frequency)
        self.inh_mini_frequency = float(inh_mini_frequency)
        self.add_params = add_params
        self.extra_attrs = {}
        cli_opts = SimConfig.cli_options
        self.local_to_global_matrix = (
            self._make_coord_map_matrix(position, rotation, scale)
            if cli_opts is None or cli_opts.enable_coord_mapping
            else False
        )

    @staticmethod
    def _make_coord_map_matrix(position, rotation, scale):
        """Build the transformation matrix from local to global"""
        if rotation is None:
            return None
        from scipy.spatial.transform import Rotation

        m = np.empty((3, 4), np.float32)
        r = Rotation.from_quat(rotation)  # scipy auto-normalizes
        m[:, :3] = r.as_matrix()
        m[:, 3] = position
        m[:, 3] *= scale
        return m

    def local_to_global_coord_mapping(self, points):
        return vector_rotate_translate(points, self.local_to_global_matrix)


def vector_rotate_translate(points, transform_matrix):
    """Rotate/translate a vector of 3D points according to a transformation matrix.

    Note: Rotation is done directly using the Einstein Sum method, similarly to scipy,
        avoiding intermediate states.
    """
    if points.shape[0] == 0:
        return np.array([])
    if len(points.shape) != 2 or points.shape[1] != 3:
        raise ValueError("Matrix of input coordinates needs 3 columns.")
    rot_matrix = transform_matrix[None, :, :3]
    translation = transform_matrix[:, 3]
    return np.einsum("ijk,ik->ij", rot_matrix, points) + translation


class METypeManager(dict):  # noqa: FURB189
    """Map to hold specific METype info and provide retrieval by gid"""

    def insert(self, gid, morph_name, *me_data, **kwargs):
        """Function to add an METypeItem to internal data structure"""
        self[int(gid)] = METypeItem(morph_name, *me_data, **kwargs)

    def load_infoNP(
        self,
        gidvec,
        morph_list,
        model_templates,
        mtypes,
        etypes,
        threshold_currents=None,
        holding_currents=None,
        exc_mini_freqs=None,
        inh_mini_freqs=None,
        positions=None,
        rotations=None,
        add_params_list=None,
    ):
        """Loads METype information in bulk from Numpy arrays"""
        for idx, gid in enumerate(gidvec):
            th_current = threshold_currents[idx] if threshold_currents is not None else 0.0
            hd_current = holding_currents[idx] if holding_currents is not None else 0.0
            exc_mini_freq = exc_mini_freqs[idx] if exc_mini_freqs is not None else 0.0
            inh_mini_freq = inh_mini_freqs[idx] if inh_mini_freqs is not None else 0.0
            position = positions[idx] if positions is not None else None
            rotation = rotations[idx] if rotations is not None else None
            mtype = mtypes[idx] if mtypes is not None else None
            add_params = add_params_list[idx] if add_params_list is not None else None
            self[int(gid)] = METypeItem(
                morph_list[idx],
                etype=etypes[idx] if etypes is not None else None,
                emodel_tpl=model_templates and model_templates[idx],
                mtype=mtype,  # TODO: check this
                threshold_current=th_current,
                holding_current=hd_current,
                exc_mini_frequency=exc_mini_freq,
                inh_mini_frequency=inh_mini_freq,
                position=position,
                rotation=rotation,
                add_params=add_params,
            )

    @property
    def gids(self):
        return self.keys()
